/* arithmetic.js - Parse simple arithmetic with Kouprey
 * Copyright (C) 2009 Rob King <jking@deadpixi.com> 
 *
 * This file is part of Kouprey.
 *
 * Kouprey is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Kouprey is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Kouprey.  If not, see <http://www.gnu.org/licenses/>.
 */

/* This file implements a simple calculator language. It supports the following
 * operations:
 *
 * Variable declaration
 *  let x = 1 + 2
 * 
 * Output
 *  print a
 *  print 1 + 2
 *
 * Four-function arithmetic
 *  1 + 2
 *  1 + 2 * 3
 *  (1 + 2) * 3
 *
 * It interprets the same grammar as arithmetic-ast.js, and also interprets it.
 */

/* Load the Kouprey library. */
load(["../../com/deadpixi/kouprey/kouprey.js"]);

/* Load the JSON serialization code. */
load(["../lib/json.js"]);

/* Load the beautifier code. */
load(["../lib/beautify.js"]);

/* Declare a dictionary for our rules. */
var Rules = {};

/* Write the grammar. */
with (com.deadpixi.kouprey) {
    with (Rules) {
        Rules.Constant      = Forward("Constant");
        Rules.Factor        = Forward("Factor");
        Rules.Term          = Forward("Term");
        Rules.Variable      = Forward("Variable");

        Rules.Space         = /^[\x20\x09\x0a\x0d]*/;
        Rules.SomeSpace     = /^[\x20\x09\x0a\x0d]+/;

        Rules.Expression    = $([Any([Space, Term, Space,
                                      $(Or("+", "-"), "operator")]),
                                      Space, Term],
                                "expression");

        Rules.Term          = $([Any([Space, Factor, Space,
                                      $(Or("*", "/"), "operator")]),
                                      Space, Factor],
                                "term");

        Rules.Factor        = $(Or(Constant,
                                   Variable,
                                   [Space, "(", Space, Expression,
                                    Space, ")"]),
                                "factor");

        Rules.Variable      = $(/^[A-Za-z_][A-Za-z0-9_]*/, "variable");

        Rules.Constant      = $(/^\-?(0|[1-9][0-9]*)(\.[0-9]+)?/, "number");

        Rules.Definition    = $(["let", SomeSpace, Variable, Space, "=",
                                 Space, Or(Constant, Variable, Expression)],
                                "definition");

        Rules.Output        = ["print", SomeSpace, $(Expression, "output")];

        Rules.Statements    = $([Some([Space,
                                       Or(Definition, Output, Expression),
                                       Space]), End()],
                                "statements");
    }
}

/* Parse the input. */
var input = File.input.readAll().join("\n");

/* Create the parser. */
var parser = new com.deadpixi.kouprey.Parser(Rules, "Statements");

/* Parse the input. */
var value = parser.parse(input);

/* Print the resulting value. */
if (value === null) {
    print("ERROR at offset " + parser.highestOffset);
} else {
    print(js_beautify(JSON.stringify(value)));
}

