/* arithmetic.js - Parse simple arithmetic with Kouprey
 * Copyright (C) 2009 Rob King <jking@deadpixi.com> 
 *
 * This file is part of Kouprey.
 *
 * Kouprey is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Kouprey is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Kouprey.  If not, see <http://www.gnu.org/licenses/>.
 */

/* This file implements a simple calculator language. It supports the following
 * operations:
 *
 * Variable declaration
 *  let x = 1 + 2
 * 
 * Output
 *  print a
 *  print 1 + 2
 *
 * Four-function arithmetic
 *  1 + 2
 *  1 + 2 * 3
 *  (1 + 2) * 3
 *
 * It interprets the same grammar as arithmetic-ast.js, and also interprets it.
 */

/* Load the Kouprey library. */
load(["../../com/deadpixi/kouprey/kouprey.js"]);

/* Declare a dictionary for our rules. */
var Rules = {};

/* Write the grammar. */
with (com.deadpixi.kouprey) {
    with (Rules) {
        Rules.Constant      = Forward("Constant");
        Rules.Factor        = Forward("Factor");
        Rules.Term          = Forward("Term");
        Rules.Variable      = Forward("Variable");

        Rules.Space         = /^[\x20\x09\x0a\x0d]*/;
        Rules.SomeSpace     = /^[\x20\x09\x0a\x0d]+/;

        Rules.Expression    = $([Any([Space, Term, Space,
                                      $(Or("+", "-"), "operator")]),
                                      Space, Term],
                                "expression");

        Rules.Term          = $([Any([Space, Factor, Space,
                                      $(Or("*", "/"), "operator")]),
                                      Space, Factor],
                                "term");

        Rules.Factor        = $(Or(Constant,
                                   Variable,
                                   [Space, "(", Space, Expression,
                                    Space, ")"]),
                                "factor");

        Rules.Variable      = $(/^[A-Za-z_][A-Za-z0-9_]*/, "variable");

        Rules.Constant      = $(/^\-?(0|[1-9][0-9]*)(\.[0-9]+)?/, "number");

        Rules.Definition    = $(["let", SomeSpace, Variable, Space, "=",
                                 Space, Or(Constant, Variable, Expression)],
                                "definition");

        Rules.Output        = ["print", SomeSpace, $(Expression, "output")];

        Rules.Statements    = $([Some([Space,
                                       Or(Definition, Output, Expression),
                                       Space]), End()],
                                "statements");
    }
}

/* Evaluate the input. */
var variables = {};
var evaluate = function (node) {
    // Function scope variables.
    var i;      // Loop index.
    var name;   // Name of a variable.
    var op;     // Operation.
    var value;  // Value of a variable.
    var v1;     // Temporary value.
    var v2;     // Temporary value.

    switch (node.type) {
        case "statements":
            for (i = 0; i < node.children.length; i++) {
                evaluate(node.children[i]);
            }
            break;

        case "output":
            print(evaluate(node.value));
            break;

        case "definition":
            name  = node.children[0].value;
            value = evaluate(node.children[1]);
            variables[name] = value;
            break;

        case "number":
            return Number(node.value);

        case "variable":
            if (variables[node.value] === undefined) {
                print("ERROR: Undefined variable: " + node.value + 
                      " at offset " + node.offset);
                quit();
            }

            return variables[node.value];

        case "factor":
        case "term":
        case "expression":
            if (node.value !== undefined) {
                return evaluate(node.value);
            }

            if (node.children.length === 1) {
                return evaluate(node.children[0]);
            }

            op = evaluate(node.children[1]);
            v1 = evaluate(node.children[0]);
            v2 = evaluate(node.children[2]);

            if (op == "+") {
                return v1 + v2;
            }

            if (op == "-") {
                return v1 - v2;
            }

            if (op == "*") {
                return v1 * v2;
            }
    
            if (op == "/") {
                return v1 / v2;
            }

            print("ERROR: Unknown operation: '" + op + "'");
            quit();
            break;

        case "operator":
            return node.value;

        default:
            print("ERROR: Unknown node type: " + node.type);
            quit();
    }
};

/* Parse the input. */
var input = File.input.readAll().join("\n");

/* Create the parser. */
var parser = new com.deadpixi.kouprey.Parser(Rules, "Statements");

/* Parse the input. */
var value = parser.parse(input);
if (value === null || value === undefined) {
    print("ERROR: Parse error at offset " + parser.highestOffset);
    quit();
}

/* Spit out the result. */
evaluate(value);
