/* cp-ast.js - Parse Component Pascal with Kouprey
 * Copyright (C) 2009 Rob King <jking@deadpixi.com> 
 *
 * This file is part of Kouprey.
 *
 * Kouprey is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Kouprey is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Kouprey.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Load the Kouprey library. */
load(["../../com/deadpixi/kouprey/kouprey.js"]);

/* Declare a dictionary for our rules. */
var Rules = {};

/* Write the grammar. */
with (com.deadpixi.kouprey) {
  with (Rules) {
    Rules.AddOp       = Forward("AddOp");
    Rules.CaseLabels  = Forward("CaseLabels");
    Rules.CaseStmt    = Forward("CaseStmt");
    Rules.Comment     = Forward("Comment");
    Rules.ConstDecl   = Forward("ConstDecl");
    Rules.ConstExpr   = Forward("ConstExpr");
    Rules.DeclSeq     = Forward("DeclSeq");
    Rules.Designator  = Forward("Designator");
    Rules.Element     = Forward("Element");
    Rules.Expr        = Forward("Expr");
    Rules.ExprList    = Forward("ExprList");
    Rules.Factor      = Forward("Factor");
    Rules.FieldList   = Forward("FieldList");
    Rules.FormalPars  = Forward("FormalPars");
    Rules.ForwardDecl = Forward("ForwardDecl");
    Rules.FPSection   = Forward("FPSection");
    Rules.Guard       = Forward("Guard");
    Rules.IdentDef    = Forward("IdentDef");
    Rules.IdentList   = Forward("IdentList");
    Rules.ImpList     = Forward("ImpList");
    Rules.MethAttrs   = Forward("MethAttrs");
    Rules.MulOp       = Forward("MulOp");
    Rules.ProcDecl    = Forward("ProcDecl");
    Rules.Qualident   = Forward("Qualident");
    Rules.Receiver    = Forward("Receiver");
    Rules.Relation    = Forward("Relation");
    Rules.Set         = Forward("Set");
    Rules.SimpleExpr  = Forward("SimpleExpr");
    Rules.Statement   = Forward("Statement");
    Rules.StmtSeq     = Forward("StmtSeq");
    Rules.Term        = Forward("Term");
    Rules.TypeDecl    = Forward("TypeDecl");
    Rules.Type        = Forward("Type");
    Rules.VarDecl     = Forward("VarDecl");

    Rules._           = Or(Some(Comment), /^\s*/);
    Rules.$_          = Or(Some(Comment), /^\s+/);

    Rules.Ident       = [Not([Or("ABSTRACT", "ARRAY", "BEGIN", "BY", "CASE",
                                 "CLOSE", "CONST", "DIV", "DO", "ELSE",
                                 "ELSIF", "EMPTY", "END", "EXIT", "EXTENSIBLE",
                                 "FOR", "IF", "IMPORT", "IN", "IS", "LIMITED",
                                 "LOOP", "MOD", "MODULE", "NIL", "OF", "OR",
                                 "OUT", "POINTER", "PROCEDURE", "RECORD",
                                 "REPEAT", "RETURN", "THEN", "TO", "TYPE",
                                 "UNTIL", "VAR", "WHILE", "WITH"), 
                                 Not(/^[A-Za-z0-9_]/)]),
                         /^[A-Za-z_][A-Za-z0-9_]*/];

    Rules.Integer     = Or(/^[0-9][0-9A-Fa-f]*[HL]/, /^[0-9]+/);
    Rules.Real        = /^[0-9]+\.[0-9]*(E[+\-]?[0-9]+)?/;
    Rules.Num         = Or(Integer, Real);

    Rules.Character   = /^[0-9][0-9A-Fa-f]*X/;

    Rules.Str         = Or(["'", /^[^']*/, "'"],
                           ['"', /^[^"]*/, '"']);

    Rules.CommentB    = "(*";
    Rules.CommentE    = "*)";
    Rules.CommentBody = [Not(CommentE), Or(Comment, /^(\s|.)/)];
    Rules.Comment     = [/^\s*/, CommentB, Any(CommentBody), CommentE, /^\s*/];

    Rules.Module      = [_, "MODULE", $_, Ident, _, ";", _,
                         Optional(ImpList), _,
                         DeclSeq, _,
                         Optional(["BEGIN", $_, StmtSeq]),
                         Optional(["CLOSE", $_, StmtSeq]), _,
                         "END", _, Ident, _, "."];

    Rules.ImpList     = ["IMPORT", _,
                         Optional([Ident, _, ":="]), _,
                         Ident, _,
                         Any([_, ",", _,
                              Optional([Ident, _, ":="]), _,
                              Ident]), _, ";"];

    Rules.DeclSeq     = [Any(Or(["CONST", $_, Any([_, ConstDecl, _, ";", _])],
                                ["TYPE", $_, Any([_, TypeDecl, _, ";", _])],
                                ["VAR", $_, Any([_, VarDecl, _, ";", _])])), _,
                         Any(Or([ProcDecl, _, ";", _],
                                [ForwardDecl, _, ";", _]))];

    Rules.ConstDecl   = [IdentDef, _, "=", _, ConstExpr];

    Rules.TypeDecl    = [IdentDef, _, "=", _, Type];

    Rules.VarDecl     = [IdentList, _, ":", _, Type];

    Rules.ProcDecl    = ["PROCEDURE", $_, 
                         Optional(Receiver), _,
                         IdentDef, _,
                         Optional(FormalPars), _,
                         MethAttrs, _,
                         Optional([";", _, DeclSeq, _,
                                   Optional(["BEGIN", $_, StmtSeq]), _,
                                   "END", $_, Ident])];
  
    Rules.MethAttrs   = [Optional([",", _, "NEW"]), _,
                         Optional([",", _, Or("ABSTRACT", "EMPTY",
                                              "EXTENSIBLE")])];

    Rules.ForwardDecl = ["PROCEDURE", _, "^", $_,
                         Optional(Receiver), _,
                         IdentDef, _,
                         Optional(FormalPars), _,
                         MethAttrs];

    Rules.FormalPars  = ["(", _, 
                         Optional([FPSection, _,
                                   Any([_, ";", _, FPSection])]), _,
                         ")", _, 
                         Optional([":", _, Type])];

    Rules.FPSection   = [Optional([Or("VAR", "IN", "OUT"), $_]), Ident, _,
                         Any([_, ",", _, Ident, _]), _,
                         ":", _, Type];

    Rules.Receiver    = ["(", _, Optional([Or("VAR", "IN"), $_]),
                         Ident, _, ":", _, Ident, _, ")"];

    Rules.Type        = Or(["ARRAY", _, 
                            Optional([ConstExpr, _, 
                                      Any([_, ",", _, ConstExpr])]), _,
                            "OF", $_, Type],

                           [Optional([Or("ABSTRACT", "EXTENSIBLE", 
                                         "LIMITED"), $_]),
                            "RECORD", $_,
                            Optional(["(", _, Qualident, _, ")"]), _,
                            FieldList, _, 
                            Any([_, ";", _, FieldList]), _,
                            "END"],

                           ["POINTER", $_, "TO", $_, Type],

                           ["PROCEDURE", _, Optional(FormalPars)],
                
                           Qualident);

    Rules.FieldList   = Optional([IdentList, _, ":", _, Type]);

    Rules.StmtSeq     = [Statement, _, Any([_, ";", _, Statement, _]), _];

    Rules.IfStmt      = ["IF", _, Expr, _, "THEN", _, StmtSeq, _,
                         Any([_, "ELSIF", _, Expr, _, "THEN", _, StmtSeq]),
                         Optional(["ELSE", _, StmtSeq]),
                         "END"];

    Rules.WhileStmt   = ["WHILE", _, Expr, _, "DO", _, StmtSeq, _, "END"];

    Rules.RepeatStmt  = ["REPEAT", _, StmtSeq, _, "UNTIL", _, Expr];

    Rules.ForStmt     = ["FOR", $_, Ident, _, ":=", _, Expr, _, "TO", _,
                         Expr, _, Optional([_, "BY", _, ConstExpr]), _,
                         "DO", _, StmtSeq, _, "END"];

    Rules.LoopStmt    = ["LOOP", $_, StmtSeq, $_, "END"];

    Rules.WithStmt    = ["WITH", _,
                         Optional([Guard, _, "DO", _, StmtSeq]), _,
                         Any(["|", _, Optional([Guard, _, "DO", _, 
                                                StmtSeq]), _]),
                         Optional(["ELSE", _, StmtSeq]), _, "END"];

    Rules.Statement   = Optional(Or(IfStmt, CaseStmt, WhileStmt, RepeatStmt,
                                    ForStmt, LoopStmt, WithStmt, "EXIT",
                                    ["RETURN", Optional([$_, Expr])],
                                    [Designator, _, ":=", _, Expr, _],
                                    [Designator, _,
                                     Optional(["(", _, Optional(ExprList), _,
                                               ")"]), _])); 

    Rules.CaseBranch  = [CaseLabels, _,
                         Any([_, ",", _, CaseLabels, _]), _,
                         ":", _, StmtSeq, _];

    Rules.CaseLabels  = [ConstExpr, _, Optional(["..", _, ConstExpr]), _];


    Rules.CaseStmt    = ["CASE", _, Expr, _, "OF", _, Optional("|"), _, 
                         CaseBranch, _,
                         Any([_, "|", _, CaseBranch, _]), _,
                         Optional(["ELSE", _, StmtSeq]), _, "END"];

    Rules.Guard       = [Qualident, _, ":", _, Qualident];

    Rules.ConstExpr   = Expr;

    Rules.Expr        = [SimpleExpr, _, Optional([Relation, _, SimpleExpr])];

    Rules.SimpleExpr  = [Optional(Or("+", "-")), _, Term, 
                         Any([_, AddOp, _, Term, _])];

    Rules.Term        = [Factor, _,
                         Any([_, MulOp, _, Factor, _]), _];

    Rules.Factor      = Or(Num, Character, Str, "NIL", Set,
                           ["(", _, Expr, _, ")"], ["~", _, Factor], 
                           Designator);

    Rules.Set         = ["{", _,
                         Optional([_, Element, _,
                                   Any([_, ",", _, Element, _])]), _,
                         "}"];

    Rules.Element     = [Expr, _, Optional(["..", _, Expr]), _];

    Rules.Relation    = Or("=", "#", "<=", ">=", "<", ">", "IN", "IS");

    Rules.AddOp       = Or("+", "-", "OR");

    Rules.MulOp       = Or("*", "/", "DIV", "MOD", "&");

    Rules.Designator  = [Qualident, _,
                         Any(Or([_, [".", Not(".")], _, Ident],
                                [_, "[", _, ExprList, _, "]"],
                                [_, "^", _],
                                [_, "(", _, Qualident, _, ")", _],
                                [_, "(", _, Optional(ExprList), _, ")", _])),
                         _, Optional("$")];

    Rules.ExprList    = [Expr, _,
                         Any([_, ",", _, Expr, _]), _];

    Rules.IdentList   = [IdentDef, _,
                         Any([_, ",", _, IdentDef, _]), _];

    Rules.Qualident   = [Optional([Ident, _, [".", Not(".")],  _]), _, Ident, _];

    Rules.IdentDef    = [Ident, _, Optional(Or("*", "-")), _];
  }
}

/* Slurp our input. */
var input = File.input.readAll().join("\n");

/* Create the parser. */
var parser = new com.deadpixi.kouprey.Parser(Rules, "Module", true);

/* Parse it. */
var output = parser.parse(input);

/* Spit out the output. */
if (output === null) {
  print("ERROR: Parsing failed at offset " + parser.highestOffset);
  print("Debug trace: ");
  parser.printDebugTrace();
} else {
    print(output);
}
