/* json.js - Parse JSON With Kouprey
 * Copyright (C) 2009 Rob King <jking@deadpixi.com> 
 *
 * This file is part of Kouprey.
 *
 * Kouprey is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Kouprey is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Kouprey.  If not, see <http://www.gnu.org/licenses/>.
 */

/* This file is a reference implementation for a JSON parser. It is a
 * direct translation of the railroad diagrams available at
 * http://json.org to a BNF-style Kouprey grammar. 
 *
 * Note that, while this is certainly elegant and shows the viability of
 * Kouprey as a parsing library, it is not particularly efficient. A faster
 * version of this parser is available in json-efficient.js.
 *
 * Note that this is strictly just a recognizer for JSON; it simply returns
 * the parsed string if the string matches the grammar.
 *
 * For a version of the JSON parser that parses JSON and deserializes it into
 * JavaScript objects, see json-full.js.
 */

/* Load the Kouprey library. */
load(["../../com/deadpixi/kouprey/kouprey.js"]);

/* Load the JSON library. Note that we just use this to print out the
 * generated AST; we don't rely on it for parsing. */
load(["../lib/json.js"]);

/* Load the beautifier library. */
load(["../lib/beautify.js"]);

/* Declare a dictionary for our rules. */
var Rules = {};

/* Write the grammar. */
with (com.deadpixi.kouprey) {
    with (Rules) {
        /* Forward declarations of the Obj and Arr rules. */
        Rules.Obj   = Forward("Obj");
        Rules.Arr   = Forward("Arr");

        /* The whitespace rule. Notice that it matches an arbitrary
         * amount of whitespace. */
        Rules.Space = /^[\x20\x09\x0d\x0a]*/;
       
        /* Match numbers:
         *  - An optional negative sign.
         *  - Either a single zero (for fractional numbers less than 1) or
         *    a leading nonzero digit followed by any number of other digits.
         *  - An optional decimal point and decimal part.
         *  - An optional exponent part consisting of either 'e' or 'E'
         *    followed by an optional sign and some digits.
         */
        Rules.Num   = [Optional("-"),                  
                       Or("0",                        
                          [/^[1-9]/, Any(/^[0-9]/)]),
                       Optional([".", Some(/^[0-9]/)]),
                       Optional([Or("e", "E"),
                                 Optional(Or("+", "-")),
                                 Some(/^[0-9]/)])];
        
        /* Match strings:
         *  - A beginning quote.
         *  - Either a backslash followed by a special character or Unicode
         *    escape sequence or
         *  - Any other character. Note that escapes are handled properly
         *    because Or conditions are processed in order of priority.
         *  - A close quote.
         */
        Rules.Str   = ['"',
                       Any(Or(["\\", Or('"', "\\", "/", "b", "f", "n", "r",
                                        "t", ["u", /^[a-fA-F0-9]{4}/])],
                              Any(/^[^\\"]/))),
                       '"'];

        /* Match a value:
         *  - A string or
         *  - A number or
         *  - An object or
         *  - An array or
         *  - One of the constants true, false, or null.
         */
        Rules.Val  = Or(Str, Num, Obj, Arr, "true", "false", "null");

        /* Match an array:
         *  - An opening bracket.
         *  - An arbitrary amount of space.
         *  - An optional array part consisting of
         *      - An arbitrary number of values followed by commas and
         *      - A final value not followed by a comma
         *  - A closing bracket.
         */
        Rules.Arr   = ["[", Space,
                       Optional([Any([Space, Val, Space, ","]),
                                 Space, Val]),
                       Space, "]"];

        /* Match an object:
         *  - An opening brace.
         *  - An arbitrary amount of space.
         *  - An optional property part consisting of
         *      - An arbitrary number of strings followed by ":" followed by
         *        a value and a comma
         *      - A final value not followed by a comma
         *  - A closing brace.
         */
        Rules.Obj   = ["{", Space,
                       Optional([Any([Space, Str, Space, ":", Space, Val,
                                      Space, ",", Space]),
                                 Space, Str, Space, ":", Space, Val]),
                       Space, "}"];

        /* A JSON text is a serialized object or array (RFC4627). */
        Rules.JSON  = Or(Obj, Arr);
    }
}

/* Parse the input. */
var input = File.input.readAll().join("\n");

/* Create the parser. */
var parser = new com.deadpixi.kouprey.Parser(Rules, "JSON");

/* Parse the input. */
var value = parser.parse(input);

/* Print the resulting value. */
if (value === null) {
    print("ERROR at offset " + parser.highestOffset);
} else {
    print(js_beautify(JSON.stringify(value)));
}
