/* json.js - Parse JSON with Kouprey
 * Copyright (C) 2009 Rob King <jking@deadpixi.com> 
 *
 * This file is part of Kouprey.
 *
 * Kouprey is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Kouprey is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with Kouprey.  If not, see <http://www.gnu.org/licenses/>.
 */

/* This file implements a full JSON deserialization system, using Kouprey to
 * parse the input JSON. */

/* Declare a dictionary for our rules. */
var Rules = {};

/* Write the grammar. */
with (com.deadpixi.kouprey) {
    with (Rules) {
        Rules.Obj   = Forward("Obj");
        Rules.Arr   = Forward("Arr");

        Rules.Space = /^[\x20\x09\x0d\x0a]*/;
        
        Rules.Num   = $(/^\-?(0|[1-9][0-9]*)(\.[0-9]+)?([Ee][+\-]?[0-9]+)?/,
                        "number");

        Rules.Chr   = Or($(/^\\(["\\/bfnrt]|u[0-9A-Fa-f]{4})/, "char"),
                         $(/^[^"\\]*/, "char"));

        Rules.Str   = ['"', $(Any(Chr), "string"), '"'];

        Rules.Val   = Or(Str, Num, $("true", "bool"), $("false", "bool"), 
                         $("null"), Obj, Arr);

        Rules.Arr   = $(["[", Space,
                         Optional([Any([Space, Val, Space, ","]),
                                   Space, Val]),
                         Space, "]"], "array");

        Rules.Pair  = $([Space, Str, Space, ":", Space, Val, Space],
                        "pair");

        Rules.Obj   = $(["{",Optional([Any([Space, Pair, Space, ","]),
                                    Pair]), "}"], "object");

        Rules.JSON  = $(Or(Obj, Arr), "json");
    }
}

/* Walk the AST and produce the values. */
var deserialize = function (node) {
    // Function scope variables.
    var a;  // Deserialized array.
    var i;  // Loop index.
    var k;  // Deserialized key.
    var o;  // Deserialized object.
    var s;  // Deserialized string.
    var v;  // Deserialized value.

    // Take appropriate action based on the type of node.
    switch (node.type) {
        case "json":
            return deserialize(node.value);

        case "object":
            o = {};
            for (i = 0; i < node.children.length; i++) {
                k = deserialize(node.children[i].children[0]);
                v = deserialize(node.children[i].children[1]);
                o[k] = v;
            }
            return o;

        case "string":
            s = "";
            for (i = 0; i < node.children.length; i++) {
                if (node.children[i].value.substring(0, 2) == "\\u") {
                    s = s + String.fromCharCode(
                        Number(node.children[i].value.substring(2, 6)));
                } else if (node.children[i].value.charAt(0) == "\\") {
                    switch (node.children[i].value.charAt(0)) {
                        case "\\":
                        case "/":
                            s = s + node.children[i].value.charAt(0);
                            break;

                        case "b":
                            s = s + "\b";
                            break;

                        case "f":
                            s = s + "\f";
                            break;

                        case "n":
                            s = s + "\n";
                            break;

                        case "r":
                            s = s + "\r";
                            break;

                        case "t":
                            s = s + "\t";
                            break;

                        default:
                            document.getElementById("error").value = 
                                ("ERROR: Invalid escape sequence " +
                                  node.children[i].value.charAt(0) + " at " +
                                  node.offset);
                    }
                } else {
                    s = s + node.children[i].value;
                }
            }
            return s;

        case "bool":
            return (node.value == "true") ? true : false;
        
        case "null":
            return null;

        case "number":
            return Number(node.value);

        case "array":
            a = [];
            for (i = 0; i < node.children.length; i++) {
                a.push(deserialize(node.children[i]));
            }
            return a;
    
        default:
            document.getElementById("error").value = 
                ("ERROR: Unknown node type: " + node.type);
            break;
    }
};

/* Create the parser. */
var parser = new com.deadpixi.kouprey.Parser(Rules, "JSON", true);

/* The parse function. */
var Run = function () {
    /* Clear the output and errror. */
    document.getElementById("output").value = "";
    document.getElementById("ast").value = "";
    document.getElementById("error").value = "";

    /* Parse the input. */
    var input = document.getElementById("input").value;

    /* Parse the input. */
    var value = parser.parse(input);

    /* Did we get an error? */
    if (value === null) {
        document.getElementById("error").value = 
            ("ERROR: Parsing failed at offset " + parser.highestOffset) + 
             "\n" + parser.getDebugTrace();
    } else {
        /* Serialize the deserialized output, to show that it worked. */
        document.getElementById("output").value = 
            (js_beautify(JSON.stringify(deserialize(value))));
    }

    /* Output the AST. */
    document.getElementById("ast").value = js_beautify(JSON.stringify(value));
}
